@@ -16,11 +16,21 @@
  */
 
 #include <fnmatch.h>
+#include <ctype.h>
+#include <dirent.h>
+#include <errno.h>
+#include <limits.h>
+#include <stdbool.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+#include <sys/stat.h>
 
 #include "queue.h"
 #include "librc.h"
 #include "helpers.h"
+#include "misc.h"
 
 bool
 rc_yesno(const char *value)
@@ -117,7 +127,7 @@ rc_getline(char **line, size_t *len, FILE *fp)
 }
 
 char *
-rc_proc_getent(const char *ent _unused)
+rc_proc_getent(const char *ent RC_UNUSED)
 {
 #ifdef __linux__
 	FILE *fp;
@@ -277,6 +287,7 @@ static RC_STRINGLIST *rc_config_kcl(RC_STRINGLIST *config)
 	overrides = rc_stringlist_new();
 
 	/* A list of variables which may be overridden on the kernel command line */
+	rc_stringlist_add(overrides, "rc_interactive");
 	rc_stringlist_add(overrides, "rc_parallel");
 
 	TAILQ_FOREACH(override, overrides, entries) {
