@@ -15,28 +15,26 @@
  *    except according to the terms contained in the LICENSE file.
  */
 
-#include <sys/types.h>
-#include <sys/ioctl.h>
-#include <sys/file.h>
-#include <sys/param.h>
-#include <sys/stat.h>
-#include <sys/wait.h>
-
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <fnmatch.h>
 #include <getopt.h>
 #include <libgen.h>
-#include <limits.h>
 #include <poll.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/ioctl.h>
+#include <sys/file.h>
+#include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 #include <termios.h>
 #include <time.h>
 #include <unistd.h>
+#include <stdbool.h>
 
 #if defined(__linux__) || (defined(__FreeBSD_kernel__) && defined(__GLIBC__)) \
 	|| defined(__GNU__)
@@ -54,6 +52,7 @@
 #include "plugin.h"
 #include "selinux.h"
 #include "_usage.h"
+#include "helpers.h"
 
 #define PREFIX_LOCK	RC_SVCDIR "/prefix.lock"
 
@@ -62,7 +61,7 @@
 #define WARN_TIMEOUT	10		/* warn about this every N seconds */
 
 const char *applet = NULL;
-const char *extraopts = "stop | start | restart | describe | zap";
+const char *extraopts = "stop | start | restart | status | describe | zap";
 const char getoptstring[] = "dDsSvl:Z" getoptstring_COMMON;
 const struct option longopts[] = {
 	{ "debug",      0, NULL, 'd'},
@@ -108,8 +107,9 @@ static RC_STRINGLIST *deptypes_mwua;	/* need+want+use+after deps for stopping */
 static void
 handle_signal(int sig)
 {
-	int serrno = errno;
-	char *signame = NULL;
+	int serrno = errno, status;
+	pid_t pid;
+	const char *signame = NULL;
 	struct winsize ws;
 
 	switch (sig) {
@@ -118,12 +118,13 @@ handle_signal(int sig)
 		break;
 
 	case SIGCHLD:
-		if (signal_pipe[1] > -1) {
-			if (write(signal_pipe[1], &sig, sizeof(sig)) == -1)
-				eerror("%s: send: %s",
-				    service, strerror(errno));
-		} else
-			rc_waitpid(-1);
+		while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
+			if (signal_pipe[1] > -1 && pid == service_pid) {
+				if (write(signal_pipe[1], &status, sizeof(status)) == -1)
+					eerror("%s: send: %s",
+					    service, strerror(errno));
+			}
+		}
 		break;
 
 	case SIGWINCH:
@@ -135,20 +136,19 @@ handle_signal(int sig)
 
 	case SIGINT:
 		if (!signame)
-			xasprintf(&signame, "SIGINT");
+			signame = "SIGINT";
 		/* FALLTHROUGH */
 	case SIGTERM:
 		if (!signame)
-			xasprintf(&signame, "SIGTERM");
+			signame = "SIGTERM";
 		/* FALLTHROUGH */
 	case SIGQUIT:
 		if (!signame)
-			xasprintf(&signame, "SIGQUIT");
+			signame = "SIGQUIT";
 		/* Send the signal to our children too */
 		if (service_pid > 0)
 			kill(service_pid, sig);
 		eerror("%s: caught %s, aborting", applet, signame);
-		free(signame);
 		exit(EXIT_FAILURE);
 		/* NOTREACHED */
 
@@ -161,7 +161,7 @@ handle_signal(int sig)
 }
 
 static void
-unhotplug()
+unhotplug(void)
 {
 	char *file = NULL;
 
@@ -440,6 +440,7 @@ svc_exec(const char *arg1, const char *arg2)
 			if (errno != EINTR) {
 				eerror("%s: poll: %s",
 				    service, strerror(errno));
+				ret = -1;
 				break;
 			}
 		}
@@ -450,9 +451,20 @@ svc_exec(const char *arg1, const char *arg2)
 				write_prefix(buffer, bytes, &prefixed);
 			}
 
-			/* Only SIGCHLD signals come down this pipe */
-			if (fd[0].revents & (POLLIN | POLLHUP))
+			/* signal_pipe receives service_pid's exit status */
+			if (fd[0].revents & (POLLIN | POLLHUP)) {
+				if ((s = read(signal_pipe[0], &ret, sizeof(ret))) != sizeof(ret)) {
+					eerror("%s: receive failed: %s", service,
+						s < 0 ? strerror(errno) : "short read");
+					ret = -1;
+					break;
+				}
+				ret = WEXITSTATUS(ret);
+				if (ret != 0 && errno == ECHILD)
+					/* killall5 -9 could cause this */
+					ret = 0;
 				break;
+			}
 		}
 	}
 
@@ -475,11 +487,6 @@ svc_exec(const char *arg1, const char *arg2)
 		master_tty = -1;
 	}
 
-	ret = rc_waitpid(service_pid);
-	ret = WEXITSTATUS(ret);
-	if (ret != 0 && errno == ECHILD)
-		/* killall5 -9 could cause this */
-		ret = 0;
 	service_pid = 0;
 
 	return ret;
@@ -614,7 +621,7 @@ svc_start_check(void)
 	}
 
 	if (exclusive_fd == -1)
-		exclusive_fd = svc_lock(applet);
+		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
 		if (errno == EACCES)
 			eerrorx("%s: superuser access required", applet);
@@ -645,7 +652,7 @@ svc_start_deps(void)
 	bool first;
 	RC_STRING *svc, *svc2;
 	RC_SERVICE state;
-	int depoptions = RC_DEP_TRACE, n;
+	int depoptions = RC_DEP_TRACE;
 	size_t len;
 	char *p, *tmp;
 	pid_t pid;
@@ -755,7 +762,6 @@ svc_start_deps(void)
 		rc_stringlist_free(use_services);
 		use_services = NULL;
 		len = 0;
-		n = 0;
 		TAILQ_FOREACH(svc, tmplist, entries) {
 			rc_service_schedule_start(svc->value, service);
 			use_services = rc_deptree_depend(deptree,
@@ -765,7 +771,6 @@ svc_start_deps(void)
 			rc_stringlist_free(use_services);
 			use_services = NULL;
 			len += strlen(svc->value) + 2;
-			n++;
 		}
 
 		len += 5;
@@ -788,7 +793,7 @@ svc_start_deps(void)
 	services = NULL;
 }
 
-static void svc_start_real()
+static void svc_start_real(void)
 {
 	bool started;
 	RC_STRING *svc, *svc2;
@@ -868,7 +873,7 @@ svc_stop_check(RC_SERVICE *state)
 		exit(EXIT_FAILURE);
 
 	if (exclusive_fd == -1)
-		exclusive_fd = svc_lock(applet);
+		exclusive_fd = svc_lock(applet, !deps);
 	if (exclusive_fd == -1) {
 		if (errno == EACCES)
 			eerrorx("%s: superuser access required", applet);
@@ -1045,26 +1050,6 @@ svc_stop(void)
 static void
 svc_restart(void)
 {
-	/* This is hairy and a better way needs to be found I think!
-	 * The issue is this - openvpn need net and dns. net can restart
-	 * dns via resolvconf, so you could have openvpn trying to restart
-	 * dnsmasq which in turn is waiting on net which in turn is waiting
-	 * on dnsmasq.
-	 * The work around is for resolvconf to restart its services with
-	 * --nodeps which means just that.
-	 * The downside is that there is a small window when our status is
-	 * invalid.
-	 * One workaround would be to introduce a new status,
-	 * or status locking. */
-	if (!deps) {
-		RC_SERVICE state = rc_service_state(service);
-		if (state & RC_SERVICE_STARTED || state & RC_SERVICE_INACTIVE)
-			svc_exec("stop", "start");
-		else
-			svc_exec("start", NULL);
-		return;
-	}
-
 	if (!(rc_service_state(service) & RC_SERVICE_STOPPED)) {
 		get_started_services();
 		svc_stop();
@@ -1112,7 +1097,6 @@ service_plugable(void)
 int main(int argc, char **argv)
 {
 	bool doneone = false;
-	bool runscript = false;
 	int retval, opt, depoptions = RC_DEP_TRACE;
 	RC_STRING *svc;
 	char *path = NULL;
@@ -1130,8 +1114,6 @@ int main(int argc, char **argv)
 	}
 
 	applet = basename_c(argv[0]);
-	if (strcmp(applet, "runscript") == 0)
-		runscript = true;
 
 	if (stat(argv[1], &stbuf) != 0) {
 		fprintf(stderr, "openrc-run `%s': %s\n",
@@ -1165,7 +1147,6 @@ int main(int argc, char **argv)
 			dir = save;
 		} else
 			file = basename_c(argv[1]);
-		ll = strlen(dir) + strlen(file) + 2;
 		xasprintf(&service, "%s/%s", dir, file);
 		if (stat(service, &stbuf) != 0) {
 			free(service);
@@ -1305,9 +1286,6 @@ int main(int argc, char **argv)
 	applet_list = rc_stringlist_new();
 	rc_stringlist_add(applet_list, applet);
 
-	if (runscript)
-		ewarn("%s uses runscript, please convert to openrc-run.", service);
-
 	/* Now run each option */
 	retval = EXIT_SUCCESS;
 	while (optind < argc) {
